/**
 ** Simple entropy harvester based upon the havege RNG
 **
 ** Copyright 2009 Gary Wuertz gary@issiweb.com
 **
 ** This program is free software: you can redistribute it and/or modify
 ** it under the terms of the GNU General Public License as published by
 ** the Free Software Foundation, either version 3 of the License, or
 ** (at your option) any later version.
 **
 ** This program is distributed in the hope that it will be useful,
 ** but WITHOUT ANY WARRANTY; without even the implied warranty of
 ** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 ** GNU General Public License for more details.
 **
 ** You should have received a copy of the GNU General Public License
 ** along with this program.  If not, see <http://www.gnu.org/licenses/>.
 **
 ** This source is an adaptation of work released as
 **
 ** Copyright (C) 2006 - André Seznec - Olivier Rochecouste
 **
 ** under version 2.1 of the GNU Lesser General Public License
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include "havege.h"
#include "havegedef.h"
/**
 * Local prototypes
 */
static void collect_ndrand(struct hperf * perf);
static int configure_ndrand(int icache,int dcache);
static int configure_hw();
static int configure_amd();
static int configure_intel();
/**
 * The original HAVEGE implementation relied heavily on compilation details
 * to tailor the mechanism to its environment. HAVEGE depends on hardware
 * details: the architecture (x96, sparc, ppc, ia64), cache sizes, and
 * the size of compiled code used in the collection sequence.
 *
 * The architectural dependence cannot be removed (since inline assembly is used),
 * but the other requirements can be moved from the compile time realm to
 * execution time to gain some additional flexibility.
 */
static struct hinfo info;
static int andpt           = 0;
static int havege_ndpt     = 0;
static int *havege_pwalk   = 0;
static int havege_bigarray [NDSIZECOLLECT + 16384];
/**
 * Entropy gathering.
 *
 * The collection loop is sized to nearly fill the instruction cache with a unrollled
 * loop of 32 instances of "oneiteration.h". The loop is completely deterministic -
 * timing fluctions occur only through the indirect effects of external events on
 * processor state. For the instruction cache, timing variations depend directly
 * on the effects of the eviction and recall of the collection sequence from the
 * cache as well as the indirect consequences such the state of branch predictors.
 * For the data cache, a large collection area and an intermediate walk table
 * the size of the data cache are used to exercise the virtual memory system and
 * the associated TBL. Again, timing fluctuations occur only as the result of
 * external events.
 */
static void collect_ndrand(struct hperf *perf)
{
   static int jumps[] = {
      &&loop32-&&loop32, &&loop31-&&loop32, &&loop30-&&loop32, &&loop29-&&loop32,
      &&loop28-&&loop32, &&loop27-&&loop32, &&loop26-&&loop32, &&loop25-&&loop32,
      &&loop24-&&loop32, &&loop23-&&loop32, &&loop22-&&loop32, &&loop21-&&loop32,
      &&loop20-&&loop32, &&loop19-&&loop32, &&loop18-&&loop32, &&loop17-&&loop32,
      &&loop16-&&loop32, &&loop15-&&loop32, &&loop14-&&loop32, &&loop13-&&loop32,
      &&loop12-&&loop32, &&loop11-&&loop32, &&loop10-&&loop32, &&loop09-&&loop32,
      &&loop08-&&loop32, &&loop07-&&loop32, &&loop06-&&loop32, &&loop05-&&loop32,
      &&loop04-&&loop32, &&loop03-&&loop32, &&loop02-&&loop32, &&loop01-&&loop32
      };
   static int PT, pt, PT2, pt2;
   static int havege_hardtick = 0;
   int *RESULT = havege_bigarray;
   int i,inter, PTtest;
   int *Pt0, *Pt1, *Pt2, *Pt3;
   struct timeval et0,et1;

   if (info.loop_idx<0) {
      int ovhd = &&loop - &&loop00;
      int sz = info.i_cache * 1024 - ovhd ;
      for(i=1;i<32&&jumps[i]<sz;i++)
         ;
      info.loop_idx = --i;
      info.loop_sz  = jumps[i]+ovhd;
      }
   i = 0;
   if (perf!=NULL)
      gettimeofday(&et0,NULL);
   goto loop00;
   #include "loopbody.h"
loop00:
   if (i < NDSIZECOLLECT) goto *(&&loop32+jumps[info.loop_idx]);
loop:
   if (perf!=NULL) {
      gettimeofday(&et1,NULL);
      perf->fill  = 1;
      perf->etime = (et1.tv_sec - et0.tv_sec)*1000000 + et1.tv_usec - et0.tv_usec;
      }
}
/**
 * Configure the collector
 *
 * Initialize the entropy collector by initializing static variables. A walk table
 * twice the size of the L1 data cache and aligned on a 4k boundary is allocated
 * and the various static vars are initialized.
 */
static int configure_ndrand (icache,dcache)
{
   int offs, *p;
   info.arch      = ARCH;
   info.vendor    = "";
   info.i_cache   = icache;
   info.d_cache   = dcache;
   info.loop_idx  = -1;
   info.loop_sz   =
   info.etime     = 0;
   if ((icache<1||dcache<1) && configure_hw()==0)
      return 0;

   andpt          = ((2*info.d_cache*1024)/sizeof(int))-1;
   p              = (int *) malloc((andpt + 4097)*sizeof(int));
   offs           = (int)((((long)&p[4096])&0xfff)/sizeof(int));
   havege_pwalk   = &p[4096-offs];
   return 1;
}
/**
 * Auto configuration using cpuid support
 */
#ifdef CPUID
static int configure_hw()
{
   unsigned char regs[4*sizeof(int)];
   unsigned int *p = (unsigned int *)regs;
   CPUID(0,p);
   switch(p[1]) {
      case 0x68747541:  info.vendor = "amd";       break;
      case 0x69727943:  info.vendor = "cyrix";     break;
      case 0x746e6543:  info.vendor = "centaur";   break;   // aka via
      case 0x756e6547:  info.vendor = "intel";     break;
      case 0x646f6547:  info.vendor = "natsemi";   break;
      case 0x52697365:
      case 0x65736952:  info.vendor = "rise";      break;   // now owned by sis
      case 0x20536953:  info.vendor = "sis";       break;
      }
   info.generic = 0;
   if (info.vendor=="amd" && configure_amd())
      ;
   else if (configure_intel())
      ;
   else {
      info.generic = 1;
      info.d_cache = GENERIC_DCACHE;
      info.i_cache = GENERIC_ICACHE;
      }
   return 1;
}
/**
 * Automatic configuration form amd chips
 *
 * As per AMD document 2541, April 2008
 */
static int configure_amd()
{
   unsigned char regs[4*sizeof(int)];
   unsigned int *p = (unsigned int *)regs;

   CPUID(0x80000000,p);
   if ((p[0]&15)>=5) {                       // We want the L1 info
      CPUID(0x80000005,p);
      info.d_cache   =  (p[2]>>24) & 0xff;   // l1 data cache
      info.i_cache   =  (p[3]>>24) & 0xff;   // l1 instruction cache
      return 1;
      }
   return 0;
}
/**
 * Automatic configuration for intel x86 chips
 *
 *  Technique below should work for ia64 with same lookup table is code is ajusted
 *  to account for a CPUID register file of 5? 64 bit words
 *
 */
static int configure_intel()
{
   unsigned char regs[4*sizeof(int)];
   unsigned int *p = (unsigned int *)regs;
  /**
   * As per Intel application note 485, December 2007 the following table contains
   * triples descriptor#, type (0=instruction,1=data), size (kb)
   * This table contains only L1 instruction(0), data(1), and trace(2) items.
   */
  static const int desc[] = {
      0x06, 0,  8 , // 4-way set assoc, 32 byte line size
      0x08, 0, 16 , // 4-way set assoc, 32 byte line size
      0x0a, 1,  8 , // 2 way set assoc, 32 byte line size
      0x0c, 1, 16 , // 4-way set assoc, 32 byte line size
      0x10, 1, 16 , // 4-way set assoc, 64 byte line size
      0x15, 0, 16 , // 4-way set assoc, 64 byte line size
      0x2c, 1, 32 , // 8-way set assoc, 64 byte line size
      0x30, 0, 32 , // 8-way set assoc, 64 byte line size
      0x60, 1, 16 , // 8-way set assoc, sectored cache, 64 byte line size
      0x66, 1,  8 , // 4-way set assoc, sectored cache, 64 byte line size
      0x67, 1, 16 , // 4-way set assoc, sectored cache, 64 byte line size
      0x68, 1, 32 , // 4-way set assoc, sectored cache, 64 byte line size
      0x70, 2, 12 , // 8-way set assoc
      0x71, 2, 16 , // 8-way set assoc
      0x72, 2, 32 , // 8-way set assoc
      0x73, 2, 64 , // 8-way set assoc
      0x77, 0, 16 , // 4-way set assoc, sectored cache, 64 byte line size
      0x00, 0,  0   // sentinel
   };
  int i,j,k,n,sizes[] = {0,0,0};

   CPUID(2,p);           	// Get cache info
   n = p[0]&0xff;
   for(i=0;i<n;i++) {
      for(j=0;j<4;j++)
         if (p[j] & 0x80000000) p[j] = 0;
      for(j=0;j<sizeof(regs);j++) {
         if (!regs[j]) continue;
         for(k=0;desc[k]!=0;k+=3)
            if (desc[k]==regs[j]) {
               sizes[desc[k+1]] += desc[k+2];
               break;
               }
         }
      if ((i+1)!=n)
         CPUID(2,p);       	// Get more cache info
      }
   if (sizes[0]<sizes[2])	   // pentium4 hack
      sizes[0] = sizes[2];
   if (info.i_cache<1)
      info.i_cache   = sizes[0];
   if (info.d_cache<1)
      info.d_cache   = sizes[1];
   if (info.i_cache>0 && info.d_cache>0)
      return 1;
   return 0;
}
#else
/**
 * Auto configuration for non-x86
 */
static int configure_hw()
{
   return 0;
}
#endif
/**
 * Configure the collection loop and seed the generator.
 */
int ndinit(int sz_icache, int sz_dcache, struct hperf *perf)
{
   if (configure_ndrand(sz_icache,sz_dcache)) {
      int max = MININITRAND*CRYPTOSIZECOLLECT/NDSIZECOLLECT;
      int i;

      for (i = 0; i < max; i++)
         collect_ndrand(perf);
      havege_ndpt = 0;
      return 1;
      }
   return 0;
}
/**
 * Return monitor information
 */
void  ndinfo(struct hinfo *rv)
{
   memcpy((void *)rv, (void *)&info, sizeof(struct hinfo));
}
/**
 * Main access point
 */
int ndrand(struct hperf *perf)
{
   if (andpt==0) exit(0);
   if (havege_ndpt >= NDSIZECOLLECT) {
      collect_ndrand(perf);
      havege_ndpt = 0;
      }
   return havege_bigarray[havege_ndpt++];
}
